# -*- coding: utf-8 -*-
# ===============================================
# (C) 2018 AO Kaspersky Lab. All Rights Reserved.
# ===============================================
from datetime import datetime
import hashlib
from itertools import chain
import json

import dt_DbHelper as DbHelper
import dt_DbHelperPg as DbHelperPg
import dt_Options as Options
from dt_classifier_common import is_local_ip

class Alerts:
    """Class to store data about alerts."""
    def __init__(self, scoringFromTime, scoringToTime):
        self.__scoringFromTime = scoringFromTime
        self.__scoringToTime = scoringToTime

        self.alert = {}
        self.infoShortDomain = {}
        self.infoApp = {}
        self.infoDll = {}
        self.infoRunner = {}
        self.infoConnection = {}
        self.infoUserAccount = {}
        self.alertHashes = {}

        self.alerts_total = 0
        self.alerts_test = 0
        self.alerts_silent = 0
        self.alerts_live = 0
        self.alert_rules = {}

        self.hz = {}
        self.rls = {}


        self.query = None

    def HasData(self):
        return len(self.alert) > 0

    _ALERT_HEADER = ['AlertMd5', 'Time', 'ScoringFromTime', 'ScoringToTime',
                     'Weight', 'Details', 'Comment', 'AlertTypeName',
                     'ShowAppLog', 'ShowHttpLog',  'ShowDllLog',
                     'ShowRunnerLog', 'ShowConnectionLog', 'ShowDnsLog',
                     'IsSilent', 'InfoData']

    _ALERT_HEADER_TYPES = [str, datetime, datetime, datetime,
                           int, str, str, str,
                           bool, bool, bool,
                           bool, bool, bool,
                           bool, str]

    def AddAlert(self, AlertMd5, Time, Weight, Details, Comment, AlertTypeName,
                 ShowAppLog = False, ShowHttpLog = False, ShowDllLog = False, ShowRunnerLog = False,
                 ShowConnectionLog = False, ShowDnsLog = False, IsSilent = False, is_test = False,
                 InfoData = None, rule_name =  None):
        AlertMd5 = AlertMd5.decode('hex')

        ScoringFromTime = self.__scoringFromTime
        ScoringToTime = self.__scoringToTime

        locals_ = locals()
        line = [locals_[h] for h in self._ALERT_HEADER]

        if AlertMd5 in self.alert:
            curIsSilent = self.alert[AlertMd5][self._ALERT_HEADER.index('IsSilent')]
            if curIsSilent and not IsSilent:
                self.alert[AlertMd5] = line
                self.alerts_silent -= 1
                self.alerts_live += 1
            else:
                line[self._ALERT_HEADER.index('Weight')] = max(Weight, self.alert[AlertMd5][self._ALERT_HEADER.index('Weight')])
                cur_comments = self.alert[AlertMd5][self._ALERT_HEADER.index('Comment')].split('\n ')
                if not Comment in cur_comments:
                    cur_comments.append(Comment)
                    cur_comments.sort()
                    self.alert[AlertMd5][self._ALERT_HEADER.index('Comment')] = '\n '.join(cur_comments)
                cur_details = self.alert[AlertMd5][self._ALERT_HEADER.index('Details')].split('\n ')
                if not Details in cur_details:
                    cur_details.append(Details)
                    self.alert[AlertMd5][self._ALERT_HEADER.index('Details')] = '\n '.join(cur_details)
        else:
            self.alert[AlertMd5] = line
            self.alerts_total += 1
            if is_test:
                self.alerts_test += 1

            if IsSilent:
                self.alerts_silent += 1
            else:
                self.alerts_live += 1

        if rule_name not in self.alert_rules:
            self.alert_rules[rule_name] = [0, 0]

        if IsSilent:
            self.alert_rules[rule_name][1] += 1
        else:
            self.alert_rules[rule_name][0] += 1


    _INFO_SHORT_DOMAIN_HEADER = [
        'AlertMd5', 'ComputerIpName', 'DomainShortName', 'RemoteHostName',
        'RegistrarName','OwnerName', 'CreationDate',
        'DomainShortKsnRank', 'DomainShortKsnUsers',
        'DomainNetFirstTime', 'DomainNetCompCount',
        'DomainShortPopularity', 'DataUploaded']

    _INFO_SHORT_DOMAIN_HEADER_TYPE = [str, str, str, str,
                                      str, str, datetime,
                                      int, int,
                                      datetime, int,
                                      int, long]

    _DATA_UPLOADED_HEADERS = [
        'ReqBodiesLengthHttp', 'ReqBodiesLength5kHttps',
        'ReqHeadersLength5kHttp', 'ReqHeadersLength5kHttps']

    _DATA_UPLOADED_HEADERS_TYPE = [long, long,
                                   long, long]

    def DomainKsnUsersToPop(self, ksnUsers):
        if ksnUsers is None or ksnUsers == 0:
            return 0
        if ksnUsers < 100:
            return 1
        if ksnUsers > 100000:
            return 5
        return 3

    def AddInfoShortDomain(self, AlertMd5, ComputerIpName,
                           DomainShortName, RemoteHostName, data):
        def _GetInfoShortDomainLineItem(i, name, args):
            if i < 3:
                return args[name]
            elif i < len(self._INFO_SHORT_DOMAIN_HEADER) - 2:
                if self._INFO_SHORT_DOMAIN_HEADER[i] == 'RemoteHostName':
                    return None
                else:
                    return data.GetValue(
                        data.domain,
                        [DomainShortName], self._INFO_SHORT_DOMAIN_HEADER[i])
            elif i == len(self._INFO_SHORT_DOMAIN_HEADER) - 2:
                ksnUsers = data.GetValue(
                    data.domain,  [DomainShortName], 'DomainShortKsnUsers')
                return self.DomainKsnUsersToPop(ksnUsers)
            else:
                uploaded = [
                    data.GetValue(data.compDomain,
                                  [ComputerIpName, DomainShortName], name)
                    for name in self._DATA_UPLOADED_HEADERS]
                return sum(u for u in uploaded if u is not None)

        AlertMd5 = AlertMd5.decode('hex')
        info = self.infoShortDomain.setdefault(AlertMd5, [])
        info.append([_GetInfoShortDomainLineItem(i, name, locals())
                     for i, name in enumerate(self._INFO_SHORT_DOMAIN_HEADER)])

    _INFO_APP_HEADER = [
        'AlertMd5', 'ComputerIpName', 'FileMd5', 'FilePathId', 'FileKsnUsers',
        'PathNormKsnUsers', 'FilePopularity', 'PathNormPopularity',
        'FileNetFirstTime', 'FileNetCompCount', 'VhsNetCompCount']

    _INFO_APP_HEADER_TYPE = [str, str, str, int, int,
                             int, int, int,
                             datetime, int, int]

    def KsnUsersToPop(self, ksnUsers):
        if ksnUsers is None or ksnUsers == 0:
            return 0
        if ksnUsers < 100:
            return 1
        if ksnUsers > 100000:
            return 5
        return 3

    def AddInfoApp(self, AlertMd5, ComputerIpName, FileMd5,
                   FilePathNorm, FilePathId, data):
        AlertMd5 = AlertMd5.decode('hex')
        FileMd5Str = FileMd5
        FileMd5 = FileMd5.decode('hex')
        line = [None] * len(self._INFO_APP_HEADER)

        locals_ = locals()
        for i in range(4):
            line[i] = locals_[self._INFO_APP_HEADER[i]]

        i = self._INFO_APP_HEADER.index('FileKsnUsers')
        line[i] = data.GetValue(data.file, FileMd5Str, 'UsersCount')

        i = self._INFO_APP_HEADER.index('FilePopularity')
        line[i] = self.KsnUsersToPop(data.GetValue(data.file, FileMd5Str, 'UsersCount'))

        i = self._INFO_APP_HEADER.index('PathNormKsnUsers')
        line[i] = data.GetValue(data.path, FilePathNorm, 'UsersCount')

        i = self._INFO_APP_HEADER.index('PathNormPopularity')
        line[i] = self.KsnUsersToPop(data.GetValue(data.path, FilePathNorm, 'UsersCount'))

        i = self._INFO_APP_HEADER.index('FileNetFirstTime')
        line[i] = data.GetValue(data.file, FileMd5Str, 'NetFirstTime')

        i = self._INFO_APP_HEADER.index('FileNetCompCount')
        line[i] = data.GetValue(data.file, FileMd5Str, 'NetCompCount')

        fileVhs = data.GetValue(data.file, FileMd5Str, 'VHS')

        if fileVhs is not None:
            i = self._INFO_APP_HEADER.index('VhsNetCompCount')
            line[i] = data.GetValue(data.vhs, fileVhs, 'NetCompCount')

        info = self.infoApp.setdefault(AlertMd5, [])
        info.append(line)

    def AddInfoApp_AllPath(self, AlertMd5, ComputerIpName, FileMd5,
                           FilePathNorm, data):
        dict = data.compAppAccount.get(ComputerIpName)
        if dict is not None:
            dict = dict.get((FileMd5, FilePathNorm))

        if dict is None:
            return

        added = set()
        for logon_info in dict.itervalues():
            for line in logon_info.itervalues():
                filePathId = data.GetValueInLine(data.compAppAccount, line,
                                                 'FilePathId')
                if filePathId not in added:
                    added.add(filePathId)
                    self.AddInfoApp(AlertMd5, ComputerIpName, FileMd5,
                                    FilePathNorm, filePathId, data)

    _INFO_DLL_HEADER = [
        'AlertMd5', 'ComputerIpName', 'DllFileMd5', 'DllPathId',
        'FileKsnUsers', 'PathNormKsnUsers', 'FilePopularity',
        'PathNormPopularity',  'FileNetFirstTime',
        'FileNetCompCount', 'VhsNetCompCount']

    _INFO_DLL_HEADER_TYPE = [str, str, str, int,
                             int, int, int,
                             int, datetime,
                             int, int]

    def AddInfoDll(self, AlertMd5, ComputerIpName, DllFileMd5, DllPathNorm,
                   DllPathId, data):
        AlertMd5 = AlertMd5.decode('hex')
        DllFileMd5Str = DllFileMd5
        DllFileMd5 = DllFileMd5.decode('hex')
        line = [None] * len(self._INFO_DLL_HEADER)

        locals_ = locals()
        for i in range(4):
            line[i] = locals_[self._INFO_DLL_HEADER[i]]

        i = self._INFO_DLL_HEADER.index('FileKsnUsers')
        line[i] = data.GetValue(data.file, DllFileMd5Str, 'UsersCount')

        i = self._INFO_APP_HEADER.index('FilePopularity')
        line[i] = self.KsnUsersToPop(
            data.GetValue(data.file, DllFileMd5Str, 'UsersCount'))

        i = self._INFO_DLL_HEADER.index('PathNormKsnUsers')
        line[i] = data.GetValue(data.path, DllPathNorm, 'UsersCount')

        i = self._INFO_APP_HEADER.index('PathNormPopularity')
        line[i] = self.KsnUsersToPop(data.GetValue(data.path, DllPathNorm, 'UsersCount'))

        i = self._INFO_DLL_HEADER.index('FileNetFirstTime')
        line[i] = data.GetValue(data.file, DllFileMd5Str, 'NetFirstTime')

        i = self._INFO_DLL_HEADER.index('FileNetCompCount')
        line[i] = data.GetValue(data.file, DllFileMd5Str, 'NetCompCount')

        dllFileVhs = data.GetValue(data.file, DllFileMd5Str, 'VHS')

        if dllFileVhs is not None:
            i = self._INFO_DLL_HEADER.index('VhsNetCompCount')
            line[i] = data.GetValue(data.vhs, dllFileVhs, 'NetCompCount')

        info = self.infoDll.setdefault(AlertMd5, [])
        info.append(line)

    _INFO_RUNNER_HEADER = [
        'AlertMd5', 'ComputerIpName', 'ParentFileMd5',
        'ParentFilePathId', 'ChildFileMd5', 'ChildFilePathId']

    _INFO_RUNNER_HEADER_TYPE = [str, str, str,
                                int, str, int]

    def AddInfoRunner(self, AlertMd5, ComputerIpName, ParentFileMd5,
                      ParentFilePathId, ChildFileMd5, ChildFilePathId, data):
        AlertMd5 = AlertMd5.decode('hex')
        ParentFileMd5 = ParentFileMd5.decode('hex')
        ChildFileMd5 = ChildFileMd5.decode('hex')
        locals_ = locals()
        line = [locals_[x] for x in self._INFO_RUNNER_HEADER]

        info = self.infoRunner.setdefault(AlertMd5, [])
        info.append(line)

    _INFO_CONNECTION_HEADER = ['AlertMd5', 'SourceIpComputerName',
                               'DestinationIpComputerName', 'Port']

    _INFO_CONNECTION_HEADER_TYPE = [str, str,
                                    str, int]

    def AddInfoConnection(self, AlertMd5, SourceIpComputerName,
                          DestinationIpComputerName, Port):
        AlertMd5 = AlertMd5.decode('hex')
        if SourceIpComputerName is None:
            SourceIpComputerName = ''

        if DestinationIpComputerName is None:
            DestinationIpComputerName = ''

        locals_ = locals()
        line = [locals_[x] for x in self._INFO_CONNECTION_HEADER]

        info = self.infoConnection.setdefault(AlertMd5, [])
        info.append(line)

    _INFO_USER_ACCOUNT_HEADER = [
        'AlertMd5', 'ComputerIpName',
        'AccountUserName', 'AccountUserDomain', 'AccountType', 'LogonType',
        'NetCompCount', 'NetFirstTime', 'CompFirstTime']

    _INFO_USER_ACCOUNT_HEADER_TYPE = [str, str,
                                      str, str, int, int,
                                      int, datetime, datetime]

    def AddInfoUserAccount(self, AlertMd5, ComputerIpName, AccountUserName,
                           AccountUserDomain, AccountType, LogonType, data):
        AlertMd5 = AlertMd5.decode('hex')
        line = [None] * len(self._INFO_USER_ACCOUNT_HEADER)

        locals_ = locals()
        for i in range(6):
            line[i] = locals_[self._INFO_USER_ACCOUNT_HEADER[i]]

        i = self._INFO_USER_ACCOUNT_HEADER.index('NetCompCount')

        line[i] = data.GetValue(
            data.user, (AccountUserName, AccountUserDomain), 'NetCompCount')

        i = self._INFO_USER_ACCOUNT_HEADER.index('NetFirstTime')
        line[i] = data.GetValue(
            data.user, (AccountUserName, AccountUserDomain), 'NetFirstTime')

        i = self._INFO_USER_ACCOUNT_HEADER.index('CompFirstTime')

        if data.compUser.has_key(ComputerIpName) and data.compUser[ComputerIpName].has_key((AccountUserName, AccountUserDomain)):
            user = data.compUser[ComputerIpName][(AccountUserName, AccountUserDomain)]
            date_index = data.compUser['HDR']['CompFirstTime']
            dates = list(line[date_index] for line in user.itervalues() if line[date_index] is not None)
            if len(dates)>0:
                line[i] = min(dates)

        info = self.infoUserAccount.setdefault(AlertMd5, [])
        info.append(line)

    def AddInfoUserAccount_AllAccounts(self, AlertMd5, ComputerIpName,
                                       FileMd5, FilePathNorm, data):
        user = data.compAppAccount.get(ComputerIpName)
        if user is not None:
            user = user.get((FileMd5, FilePathNorm))

        if user is None:
            return

        added = set()
        for (name, domain), account in user.iteritems():
            for account_type, logon_type in account.iterkeys():
                key = "{0}{1}{2}{3}".format(name, domain,
                                            account_type, logon_type)
                if key not in added:
                    added.add(key)
                    self.AddInfoUserAccount(
                        AlertMd5, ComputerIpName, name,  domain,
                        account_type, logon_type, data)



    _ALERT_HASHES_HEADER = [
        'AlertMd5', 'AlertHashTypeName', 'AlertHash', 'AlertHashValue']

    _ALERT_HASHES_TYPE = [str, str, str, str]

    def AddAlertHashes(self, AlertMd5, AlertHashTypeName, AlertHash, AlertHashValue = 'Empty'):
        AlertMd5 = AlertMd5.decode('hex')
        AlertHash = AlertHash.decode('hex')

        line = [None] * len(self._ALERT_HASHES_HEADER)

        locals_ = locals()
        for i in range(4):
            line[i] = locals_[self._ALERT_HASHES_HEADER[i]]

        info = self.alertHashes.setdefault(AlertMd5, [])
        info.append(line)




    def ClearAlerts(self):
        self.alert = {}
        self.infoShortDomain = {}
        self.infoApp = {}
        self.infoDll = {}
        self.infoRunner = {}
        self.infoConnection = {}
        self.infoUserAccount = {}
        self.alertHashes = {}

    _HASH_THIS = set([
                      'AccountUserName'.lower(),
                      'AccountUserDomain'.lower()])

    _HASH_FOR_LOCAL_IP = set(['ComputerIpName'.lower(),
                              'DestinationComputerIpName'.lower(),
                              'SourceComputerIpName'.lower(),
                              "DomainShortName".lower()]);

    def HexString(self, text, do_hashing=True):
        hash = hashlib.md5(text).digest() if do_hashing else text
        return "".join("{0:02x}".format(ord(c)) for c in hash)

    def HashColumns(self, lines, header):
        lines_hashed = []
        for line in lines:
            this_line_dict = {}
            for name, item in zip(header, line):
                if item is None:
                    continue

                if name.lower() in self._HASH_THIS:
                    value = self.HexString(item)
                elif name.lower() in self._HASH_FOR_LOCAL_IP and is_local_ip(item):
                    value = self.HexString(item)
                elif ('md5' in name.lower() or 'alerthash' == name.lower()) and len(item)==16:
                    value = self.HexString(item, do_hashing=False)
                else:
                    value = str(item)

                this_line_dict[name] = value

            lines_hashed.append(this_line_dict)

        return lines_hashed

    def MakeAlertsListWithInfo(self):
        result = []
        infoDataIdx = self._ALERT_HEADER.index('InfoData')
        for alert_md5, alert_data in self.alert.iteritems():
            infoData = {}

            infoData['Alert'] = self.HashColumns([alert_data], self._ALERT_HEADER)

            info = self.infoShortDomain.get(alert_md5)
            if info is not None:
                infoData['infoShortDomain'] = self.HashColumns(info, self._INFO_SHORT_DOMAIN_HEADER)

            info = self.infoApp.get(alert_md5)
            if info is not None:
                infoData['infoApp'] = self.HashColumns(info, self._INFO_APP_HEADER)

            info = self.infoDll.get(alert_md5)
            if info is not None:
                infoData['infoDll'] = self.HashColumns(info, self._INFO_DLL_HEADER)

            info = self.infoRunner.get(alert_md5)
            if info is not None:
                infoData['infoRunner'] = self.HashColumns(info, self._INFO_RUNNER_HEADER)

            info = self.infoConnection.get(alert_md5)
            if info is not None:
                infoData['infoConnection'] = self.HashColumns(info, self._INFO_CONNECTION_HEADER)

            info = self.infoUserAccount.get(alert_md5)
            if info is not None:
                infoData['infoUser'] = self.HashColumns(info,  self._INFO_USER_ACCOUNT_HEADER)

            info = self.alertHashes.get(alert_md5)
            if info is not None:
                infoData['alertHashes'] = self.HashColumns(info, self._ALERT_HASHES_HEADER)

            alert_data[infoDataIdx] = json.dumps(infoData).replace('\\', '\\\\')
            result.append(alert_data)

        return result

    def DictToListInfo(self, dict):
        return list(chain.from_iterable(dict.itervalues()))

    def SaveAlerts(self):
        if Options.AptDbType == 'PGSQL':
            dbHelper = DbHelperPg.DbHelperPg(Options.PgConnection)
            prefix = "t"
            qProc = 'select alert.AddAlerts();'
        elif Options.AptDbType == 'MSSQL':
            dbHelper = DbHelper.DbHelperMs(
                Options.MsServer, Options.MsUser, Options.MsPassword,
                Options.MsDataBase)
            prefix = "#"
            qProc = 'exec alert.AddAlerts'
        else:
            raise ValueError("Unknown AptDbType type '{0}'"
                             .format(Options.AptDbType))
        query = []

        if self.alert:
            list = self.MakeAlertsListWithInfo()
            query.append(dbHelper.ListToSqlTableTypes(
                self._ALERT_HEADER, self._ALERT_HEADER_TYPES, list,
                prefix + "Alert", ['AlertMd5']).replace(
                'InfoData text null', 'InfoData bytea null'))

        if self.infoShortDomain:
            list = self.DictToListInfo(self.infoShortDomain)
            query.append(dbHelper.ListToSqlTableTypes(
                self._INFO_SHORT_DOMAIN_HEADER,
                self._INFO_SHORT_DOMAIN_HEADER_TYPE, list,
                prefix + "InfoDomainShort", ['AlertMd5']))

        if self.infoApp:
            list = self.DictToListInfo(self.infoApp)
            query.append(dbHelper.ListToSqlTableTypes(
                self._INFO_APP_HEADER, self._INFO_APP_HEADER_TYPE, list,
                prefix + "InfoApp", ['AlertMd5', 'FileMd5']))

        if self.infoDll:
            list = self.DictToListInfo(self.infoDll)
            query.append(dbHelper.ListToSqlTableTypes(
                self._INFO_DLL_HEADER, self._INFO_DLL_HEADER_TYPE, list,
                prefix + "InfoDll", ['AlertMd5', 'DllFileMd5']))

        if self.infoRunner:
            list = self.DictToListInfo(self.infoRunner)
            query.append(dbHelper.ListToSqlTableTypes(
                self._INFO_RUNNER_HEADER, self._INFO_RUNNER_HEADER_TYPE, list,
                prefix + "InfoRunner",
                ['AlertMd5', 'ParentFileMd5', 'ChildFileMd5']))

        if self.infoConnection:
            list = self.DictToListInfo(self.infoConnection)
            query.append(dbHelper.ListToSqlTableTypes(
                self._INFO_CONNECTION_HEADER,
                self._INFO_CONNECTION_HEADER_TYPE, list,
                prefix + "InfoConnection", ['AlertMd5']))

        if self.infoUserAccount:
            list = self.DictToListInfo(self.infoUserAccount)
            query.append(dbHelper.ListToSqlTableTypes(
                self._INFO_USER_ACCOUNT_HEADER,
                self._INFO_USER_ACCOUNT_HEADER_TYPE, list,
                prefix + "InfoUserAccount", ['AlertMd5']))

        if self.alertHashes:
            list = self.DictToListInfo(self.alertHashes)
            query.append(dbHelper.ListToSqlTableTypes(
                self._ALERT_HASHES_HEADER, self._ALERT_HASHES_TYPE, list,
                prefix + "addAlertHashes", ['AlertMd5', 'AlertHash']))


        self.alert = {}
        self.infoShortDomain = {}
        self.infoApp = {}
        self.infoDll = {}
        self.infoRunner = {}
        self.infoConnection = {}
        self.infoUserAccount = {}
        self.alertHashes = {}

        if query:
            query.append(qProc)
            query = "\n\n".join(query)
            self.query = query

            dbHelper.ExecuteScalarQuery(query)
